// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2024, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#include "mo_lib_intp_rbf.hpp"
#include <Kokkos_Core.hpp>

constexpr int rbf_vec_dim_v = 6;

//-------------------------------------------------------------------------
//
//
//>
/// Performs vector RBF reconstruction at triangle vertices.
///
/// Theory described in Narcowich and Ward (Math Comp. 1994) and
/// Bonaventura and Baudisch (Mox Report n. 75).
/// It takes edge based variables as input and combines them
/// into three dimensional cartesian vectors at each vertex.
///
/// Two templated variables in order to support mixed precision.
/// Intended that type_traits::is_floating_point(T,S)==TRUE
/// precision(T) >= precision(S)
template <typename T, typename S>
void rbf_vec_interpol_vertex_lib(
    const T *p_e_in, const int *rbf_vec_idx_v, const int *rbf_vec_blk_v,
    const T *rbf_vec_coeff_v, S *p_u_out, S *p_v_out,
    const int i_startblk,    // start_block needed for get_indices_c_lib
    const int i_endblk,      // end_block needed for get_indices_c_lib
    const int i_startidx_in, // start_index needed for get_indices_c_lib
    const int i_endidx_in,   // end_index needed for get_indices_c_lib
    const int slev,          // vertical start level
    const int elev,          // vertical end level
    const int nproma,        // inner loop length/vector length
    const bool lacc,         // if true, use Cuda mem-/exec-spaces
    const bool acc_async,    // [deprecated] use async acc
    // Dimensions for the arrays.
    const int nlev, const int nblks_e, const int nblks_v) {
  /*
#ifdef DIM_ENABLE_GPU
  if (lacc){ using MemSpace = Kokkos::CudaSpace;
  } else { using MemSpace = Kokkos::HostSpace; }
#else
  using MemSpace = Kokkos::HostSpace;
#endif

  */

  // Wrap raw pointers in unmanaged Kokkos Views.
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<const T ****, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstT4D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;
  typedef Kokkos::View<S ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedS3D;

  // input components of velocity or horizontal vorticity vectors at edge
  // midpoints dim: (nproma,nlev,nblks_e)
  UnmanagedConstT3D p_e_in_view(p_e_in, nproma, nlev, nblks_e);

  // index array defining the stencil of surrounding edges for vector rbf
  // interpolation at each triangle vertex (rbf_vec_dim_v,nproma,nblks_v)
  UnmanagedConstInt3D iidx_view(rbf_vec_idx_v, rbf_vec_dim_v, nproma, nblks_v);
  UnmanagedConstInt3D iblk_view(rbf_vec_blk_v, rbf_vec_dim_v, nproma, nblks_v);

  // coefficients are working precision array containing the coefficients used
  // for vector rbf interpolation at each tringle vertex (input is normal
  // component), dim: (rbf_vec_dim_v,2,nproma,nblks_v)
  UnmanagedConstT4D ptr_coeff_view(rbf_vec_coeff_v, rbf_vec_dim_v, 2, nproma,
                                   nblks_v);

  // reconstructed x-component (u) of velocity vector,
  // dim: (nproma,nlev,nblks_v)
  UnmanagedS3D p_u_out_view(p_u_out, nproma, nlev, nblks_v);
  // reconstructed y-component (v) of velocity vector,
  // dim: (nproma,nlev,nblks_v)
  UnmanagedS3D p_v_out_view(p_v_out, nproma, nlev, nblks_v);

  // Local vars
  // int jv, jk, jb; // integer over vertices, levels, and blocks,
  int jb;         // integer over vertices, levels, and blocks,
  int i_startidx; // start index
  int i_endidx;   // end index

  for (jb = i_startblk; jb <= i_endblk; ++jb) {

    get_indices_v_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});

    Kokkos::parallel_for(
        "rbf_vec_interpol_vertex_lib", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int jv) {
          // NOTE: Static indexes reduced by 1 from Fortran version
          p_u_out_view(jv, jk, jb) =
              ptr_coeff_view(0, 0, jv, jb) *
                  p_e_in_view(iidx_view(0, jv, jb), jk, iblk_view(0, jv, jb)) +
              ptr_coeff_view(1, 0, jv, jb) *
                  p_e_in_view(iidx_view(1, jv, jb), jk, iblk_view(1, jv, jb)) +
              ptr_coeff_view(2, 0, jv, jb) *
                  p_e_in_view(iidx_view(2, jv, jb), jk, iblk_view(2, jv, jb)) +
              ptr_coeff_view(3, 0, jv, jb) *
                  p_e_in_view(iidx_view(3, jv, jb), jk, iblk_view(3, jv, jb)) +
              ptr_coeff_view(4, 0, jv, jb) *
                  p_e_in_view(iidx_view(4, jv, jb), jk, iblk_view(4, jv, jb)) +
              ptr_coeff_view(5, 0, jv, jb) *
                  p_e_in_view(iidx_view(5, jv, jb), jk, iblk_view(5, jv, jb));
          p_v_out_view(jv, jk, jb) =
              ptr_coeff_view(0, 1, jv, jb) *
                  p_e_in_view(iidx_view(0, jv, jb), jk, iblk_view(0, jv, jb)) +
              ptr_coeff_view(1, 1, jv, jb) *
                  p_e_in_view(iidx_view(1, jv, jb), jk, iblk_view(1, jv, jb)) +
              ptr_coeff_view(2, 1, jv, jb) *
                  p_e_in_view(iidx_view(2, jv, jb), jk, iblk_view(2, jv, jb)) +
              ptr_coeff_view(3, 1, jv, jb) *
                  p_e_in_view(iidx_view(3, jv, jb), jk, iblk_view(3, jv, jb)) +
              ptr_coeff_view(4, 1, jv, jb) *
                  p_e_in_view(iidx_view(4, jv, jb), jk, iblk_view(4, jv, jb)) +
              ptr_coeff_view(5, 1, jv, jb) *
                  p_e_in_view(iidx_view(5, jv, jb), jk, iblk_view(5, jv, jb));
        });
  }
}

template <typename T>
void rbf_interpol_c2grad_lib(const T *p_cell_in, const int *rbf_c2grad_idx,
                             const int *rbf_c2grad_blk,
                             const T *rbf_c2grad_coeff, T *grad_x, T *grad_y,
                             int i_startblk, int i_endblk, int i_startidx_in,
                             int i_endidx_in, int slev, int elev, int nproma,
                             int rbf_c2grad_dim, int nlev, int nblks_c,
                             bool lacc) {

  // aliases for unmanaged Kokkos views
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedT3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;
  typedef Kokkos::View<const T ****, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstT4D;

  // to avoid memory ownership issues
  UnmanagedConstT3D p_cell_in_view(p_cell_in, nproma, nlev, nblks_c);
  UnmanagedT3D grad_x_view(grad_x, nproma, nlev, nblks_c);
  UnmanagedT3D grad_y_view(grad_y, nproma, nlev, nblks_c);
  UnmanagedConstInt3D rbf_c2grad_idx_view(rbf_c2grad_idx, rbf_c2grad_dim,
                                          nproma, nblks_c);
  UnmanagedConstInt3D rbf_c2grad_blk_view(rbf_c2grad_blk, rbf_c2grad_dim,
                                          nproma, nblks_c);
  UnmanagedConstT4D rbf_c2grad_coeff_view(rbf_c2grad_coeff, rbf_c2grad_dim, 2,
                                          nproma, nblks_c);

  for (int jb = i_startblk; jb <= i_endblk; ++jb) {

    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});

    Kokkos::parallel_for(
        "rbf_interpol_c2grad", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int jc) {
          grad_x_view(jc, jk, jb) =
              rbf_c2grad_coeff_view(0, 1, jc, jb) * p_cell_in_view(jc, jk, jb) +
              rbf_c2grad_coeff_view(1, 1, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(1, jc, jb), jk,
                                 rbf_c2grad_blk_view(1, jc, jb)) +
              rbf_c2grad_coeff_view(2, 1, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(2, jc, jb), jk,
                                 rbf_c2grad_blk_view(2, jc, jb)) +
              rbf_c2grad_coeff_view(3, 1, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(3, jc, jb), jk,
                                 rbf_c2grad_blk_view(3, jc, jb)) +
              rbf_c2grad_coeff_view(4, 1, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(4, jc, jb), jk,
                                 rbf_c2grad_blk_view(4, jc, jb)) +
              rbf_c2grad_coeff_view(5, 1, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(5, jc, jb), jk,
                                 rbf_c2grad_blk_view(5, jc, jb)) +
              rbf_c2grad_coeff_view(6, 1, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(6, jc, jb), jk,
                                 rbf_c2grad_blk_view(6, jc, jb)) +
              rbf_c2grad_coeff_view(7, 1, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(7, jc, jb), jk,
                                 rbf_c2grad_blk_view(7, jc, jb)) +
              rbf_c2grad_coeff_view(8, 1, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(8, jc, jb), jk,
                                 rbf_c2grad_blk_view(8, jc, jb)) +
              rbf_c2grad_coeff_view(9, 1, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(9, jc, jb), jk,
                                 rbf_c2grad_blk_view(9, jc, jb));

          grad_y_view(jc, jk, jb) =
              rbf_c2grad_coeff_view(0, 2, jc, jb) * p_cell_in_view(jc, jk, jb) +
              rbf_c2grad_coeff_view(1, 2, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(1, jc, jb), jk,
                                 rbf_c2grad_blk_view(1, jc, jb)) +
              rbf_c2grad_coeff_view(2, 2, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(2, jc, jb), jk,
                                 rbf_c2grad_blk_view(2, jc, jb)) +
              rbf_c2grad_coeff_view(3, 2, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(3, jc, jb), jk,
                                 rbf_c2grad_blk_view(3, jc, jb)) +
              rbf_c2grad_coeff_view(4, 2, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(4, jc, jb), jk,
                                 rbf_c2grad_blk_view(4, jc, jb)) +
              rbf_c2grad_coeff_view(5, 2, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(5, jc, jb), jk,
                                 rbf_c2grad_blk_view(5, jc, jb)) +
              rbf_c2grad_coeff_view(6, 2, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(6, jc, jb), jk,
                                 rbf_c2grad_blk_view(6, jc, jb)) +
              rbf_c2grad_coeff_view(7, 2, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(7, jc, jb), jk,
                                 rbf_c2grad_blk_view(7, jc, jb)) +
              rbf_c2grad_coeff_view(8, 2, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(8, jc, jb), jk,
                                 rbf_c2grad_blk_view(8, jc, jb)) +
              rbf_c2grad_coeff_view(9, 2, jc, jb) *
                  p_cell_in_view(rbf_c2grad_idx_view(9, jc, jb), jk,
                                 rbf_c2grad_blk_view(9, jc, jb));
        });

  } // for
} // void

//------------------------------------------rbf_vec_interpol_cell_lib---------------------------------------------

template <typename T>
void rbf_vec_interpol_cell_lib(const T *p_vn_in, const int *rbf_vec_idx_c,
                               const int *rbf_vec_blk_c,
                               const T *rbf_vec_coeff_c, T *p_u_out, T *p_v_out,
                               int i_startblk, int i_endblk, int i_startidx_in,
                               int i_endidx_in, int slev, int elev, int nproma,
                               int nlev, int nblks_c, int nblks_e,
                               int rbf_vec_dim_c, bool lacc, bool acc_async) {

  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedT3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;
  typedef Kokkos::View<const T ****, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstT4D;

  UnmanagedConstT3D p_vn_in_view(p_vn_in, nproma, nlev, nblks_e);
  UnmanagedConstInt3D rbf_vec_idx_c_view(rbf_vec_idx_c, rbf_vec_dim_c, nproma,
                                         nblks_c);
  UnmanagedConstInt3D rbf_vec_blk_c_view(rbf_vec_blk_c, rbf_vec_dim_c, nproma,
                                         nblks_c);
  UnmanagedConstT4D rbf_vec_coeff_c_view(rbf_vec_coeff_c, nproma,
                                         nblks_c); // TODO
  UnmanagedT3D p_u_out_view(p_u_out, nproma, nlev, nblks_c);
  UnmanagedT3D p_v_out_view(p_u_out, nproma, nlev, nblks_c);

  for (int jb = i_startblk; jb <= i_endblk; ++jb) {

    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});

    Kokkos::parallel_for(
        "rbf_vec_interpol_cell_lib", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int jc) {
          p_u_out_view(jc, jk, jb) =
              rbf_vec_coeff_c_view(0, 1, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(0, jc, jb), jk,
                               rbf_vec_blk_c_view(0, jc, jb)) +
              rbf_vec_coeff_c_view(1, 1, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(1, jc, jb), jk,
                               rbf_vec_blk_c_view(1, jc, jb)) +
              rbf_vec_coeff_c_view(2, 1, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(2, jc, jb), jk,
                               rbf_vec_blk_c_view(2, jc, jb)) +
              rbf_vec_coeff_c_view(3, 1, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(3, jc, jb), jk,
                               rbf_vec_blk_c_view(3, jc, jb)) +
              rbf_vec_coeff_c_view(4, 1, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(4, jc, jb), jk,
                               rbf_vec_blk_c_view(4, jc, jb)) +
              rbf_vec_coeff_c_view(5, 1, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(5, jc, jb), jk,
                               rbf_vec_blk_c_view(5, jc, jb)) +
              rbf_vec_coeff_c_view(6, 1, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(6, jc, jb), jk,
                               rbf_vec_blk_c_view(6, jc, jb)) +
              rbf_vec_coeff_c_view(7, 1, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(7, jc, jb), jk,
                               rbf_vec_blk_c_view(7, jc, jb)) +
              rbf_vec_coeff_c_view(8, 1, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(8, jc, jb), jk,
                               rbf_vec_blk_c_view(8, jc, jb));

          p_v_out_view(jc, jk, jb) =
              rbf_vec_coeff_c_view(0, 2, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(0, jc, jb), jk,
                               rbf_vec_blk_c_view(0, jc, jb)) +
              rbf_vec_coeff_c_view(1, 2, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(1, jc, jb), jk,
                               rbf_vec_blk_c_view(1, jc, jb)) +
              rbf_vec_coeff_c_view(2, 2, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(2, jc, jb), jk,
                               rbf_vec_blk_c_view(2, jc, jb)) +
              rbf_vec_coeff_c_view(3, 2, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(3, jc, jb), jk,
                               rbf_vec_blk_c_view(3, jc, jb)) +
              rbf_vec_coeff_c_view(4, 2, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(4, jc, jb), jk,
                               rbf_vec_blk_c_view(4, jc, jb)) +
              rbf_vec_coeff_c_view(5, 2, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(5, jc, jb), jk,
                               rbf_vec_blk_c_view(5, jc, jb)) +
              rbf_vec_coeff_c_view(6, 2, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(6, jc, jb), jk,
                               rbf_vec_blk_c_view(6, jc, jb)) +
              rbf_vec_coeff_c_view(7, 2, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(7, jc, jb), jk,
                               rbf_vec_blk_c_view(7, jc, jb)) +
              rbf_vec_coeff_c_view(8, 2, jc, jb) *
                  p_vn_in_view(rbf_vec_idx_c_view(8, jc, jb), jk,
                               rbf_vec_blk_c_view(8, jc, jb));
        });
    Kokkos::fence();
  } // for
} // void

//------------------------------------------rbf_vec_interpol_edge_lib---------------------------------------------

template <typename T>
void rbf_vec_interpol_edge_lib(const T *p_vn_in, const int *rbf_vec_idx_e,
                               const int *rbf_vec_blk_e,
                               const T *rbf_vec_coeff_e, T *p_vt_out,
                               int i_startblk, int i_endblk, int i_startidx_in,
                               int i_endidx_in, int slev, int elev, int nlev,
                               int nproma, int rbf_vec_dim_e, int nblks_e,
                               bool lacc, bool acc_async) {

  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedT3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;

  UnmanagedConstT3D p_vn_in_view(p_vn_in, nproma, nlev, nblks_e);
  UnmanagedConstInt3D rbf_vec_idx_e_view(rbf_vec_idx_e, rbf_vec_dim_e, nproma,
                                         nblks_e);
  UnmanagedConstInt3D rbf_vec_blk_e_view(rbf_vec_blk_e, rbf_vec_dim_e, nproma,
                                         nblks_e);
  UnmanagedConstT3D rbf_vec_coeff_e_view(rbf_vec_coeff_e, rbf_vec_dim_e, nproma,
                                         nblks_e);
  UnmanagedT3D p_vt_out_view(p_vt_out, nproma, nlev, nblks_e);

  for (int jb = i_startblk; jb <= i_endblk; ++jb) {

    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});

    Kokkos::parallel_for(
        "rbf_vec_interpol_edge_lib", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int je) {
          p_vt_out_view(je, jk, jb) =
              rbf_vec_coeff_e_view(0, je, jb) *
                  p_vn_in_view(rbf_vec_idx_e_view(0, je, jb), jk,
                               rbf_vec_blk_e_view(0, je, jb)) +
              rbf_vec_coeff_e_view(1, je, jb) *
                  p_vn_in_view(rbf_vec_idx_e_view(1, je, jb), jk,
                               rbf_vec_blk_e_view(1, je, jb)) +
              rbf_vec_coeff_e_view(2, je, jb) *
                  p_vn_in_view(rbf_vec_idx_e_view(2, je, jb), jk,
                               rbf_vec_blk_e_view(2, je, jb)) +
              rbf_vec_coeff_e_view(3, je, jb) *
                  p_vn_in_view(rbf_vec_idx_e_view(3, je, jb), jk,
                               rbf_vec_blk_e_view(3, je, jb));
        });
  }
}

// Explicit instantiation - double precision
template void rbf_vec_interpol_vertex_lib<double, double>(
    const double *p_e_in, const int *rbf_vec_idx_v, const int *rbf_vec_blk_v,
    const double *rbf_vec_coeff_v, double *p_u_out, double *p_v_out,
    const int i_startblk, const int i_endblk, const int i_startidx_in,
    const int i_endidx_in, const int slev, const int elev, const int nproma,
    const bool lacc, const bool acc_async, const int nlev, const int nblks_e,
    const int nblks_v);

// Explicit instantiation - single precision
template void rbf_vec_interpol_vertex_lib<float, float>(
    const float *p_e_in, const int *rbf_vec_idx_v, const int *rbf_vec_blk_v,
    const float *rbf_vec_coeff_v, float *p_u_out, float *p_v_out,
    const int i_startblk, const int i_endblk, const int i_startidx_in,
    const int i_endidx_in, const int slev, const int elev, const int nproma,
    const bool lacc, const bool acc_async, const int nlev, const int nblks_e,
    const int nblks_v);

// Explicit instantiation - mixed precision
template void rbf_vec_interpol_vertex_lib<double, float>(
    const double *p_e_in, const int *rbf_vec_idx_v, const int *rbf_vec_blk_v,
    const double *rbf_vec_coeff_v, float *p_u_out, float *p_v_out,
    const int i_startblk, const int i_endblk, const int i_startidx_in,
    const int i_endidx_in, const int slev, const int elev, const int nproma,
    const bool lacc, const bool acc_async, const int nlev, const int nblks_e,
    const int nblks_v);

template void rbf_vec_interpol_cell_lib<double>(
    const double *p_vn_in, const int *rbf_vec_idx_c, const int *rbf_vec_blk_c,
    const double *rbf_vec_coeff_c, double *p_u_out, double *p_v_out,
    int i_startblk, int i_endblk, int i_startidx_in, int i_endidx_in, int slev,
    int elev, int nproma, int nlev, int nblks_c, int nblks_e, int rbf_vec_dim_c,
    bool lacc, bool acc_async);

template void rbf_vec_interpol_cell_lib<float>(
    const float *p_vn_in, const int *rbf_vec_idx_c, const int *rbf_vec_blk_c,
    const float *rbf_vec_coeff_c, float *p_u_out, float *p_v_out,
    int i_startblk, int i_endblk, int i_startidx_in, int i_endidx_in, int slev,
    int elev, int nproma, int nlev, int nblks_c, int nblks_e, int rbf_vec_dim_c,
    bool lacc, bool acc_async);

template void rbf_interpol_c2grad_lib<double>(
    const double *p_cell_in, const int *rbf_c2grad_idx,
    const int *rbf_c2grad_blk, const double *rbf_c2grad_coeff, double *grad_x,
    double *grad_y, int i_startblk, int i_endblk, int i_startidx_in,
    int i_endidx_in, int slev, int elev, int nproma, int rbf_c2grad_dim,
    int nlev, int nblks_c, bool lacc);

template void rbf_interpol_c2grad_lib<float>(
    const float *p_cell_in, const int *rbf_c2grad_idx,
    const int *rbf_c2grad_blk, const float *rbf_c2grad_coeff, float *grad_x,
    float *grad_y, int i_startblk, int i_endblk, int i_startidx_in,
    int i_endidx_in, int slev, int elev, int nproma, int rbf_c2grad_dim,
    int nlev, int nblks_c, bool lacc);

template void rbf_vec_interpol_edge_lib<double>(
    const double *p_vn_in, const int *rbf_vec_idx_e, const int *rbf_vec_blk_e,
    const double *rbf_vec_coeff_e, double *p_vt_out, int i_startblk,
    int i_endblk, int i_startidx_in, int i_endidx_in, int slev, int elev,
    int nlev, int nproma, int rbf_vec_dim_e, int nblks_e, bool lacc,
    bool acc_async);

template void rbf_vec_interpol_edge_lib<float>(
    const float *p_vn_in, const int *rbf_vec_idx_e, const int *rbf_vec_blk_e,
    const float *rbf_vec_coeff_e, float *p_vt_out, int i_startblk, int i_endblk,
    int i_startidx_in, int i_endidx_in, int slev, int elev, int nlev,
    int nproma, int rbf_vec_dim_e, int nblks_e, bool lacc, bool acc_async);
