// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2024, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#include "mo_math_utilities.hpp"

template <typename T>
void tdma_solver_vec(const T* a, const T* b, const T* c, const T* d,
                              int slev, int elev, int startidx, int endidx,
                              int nrows, int ncols, T* varout) {

  // Start timing
  auto start_time = std::chrono::high_resolution_clock::now();

  // Allocate temporary arrays using Kokkos::View.
  // The views c_p and d_p are allocated as 2D arrays with dimensions [nrows][ncols].
  // Kokkos::View automatically handles memory management.
  Kokkos::View<T**> c_p("c_p", nrows, ncols);
  Kokkos::View<T**> d_p("d_p", nrows, ncols);

  // Wrap the input arrays in unmanaged views.
  // We assume that the input arrays are laid out in column-major order as in the original code.
  // Here we use LayoutLeft so that the first index (row) is contiguous.
  typedef Kokkos::View<const T**, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged> UnmanagedConst2D;
  typedef Kokkos::View<T**, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged> Unmanaged2D;
  UnmanagedConst2D a_view(a, nrows, ncols);
  UnmanagedConst2D b_view(b, nrows, ncols);
  UnmanagedConst2D c_view(c, nrows, ncols);
  UnmanagedConst2D d_view(d, nrows, ncols);
  Unmanaged2D varout_view(varout, nrows, ncols);

  // Initialize c-prime and d-prime at the starting level (slev)
  Kokkos::parallel_for("init_c_p_d_p", Kokkos::RangePolicy<>(startidx, endidx), KOKKOS_LAMBDA (const int jc) {
    c_p(jc, slev) = c_view(jc, slev) / b_view(jc, slev);
    d_p(jc, slev) = d_view(jc, slev) / b_view(jc, slev);
  });
  Kokkos::fence();

  // Forward sweep: compute c-prime and d-prime for each column from slev+1 to elev-1.
  for (int i = slev + 1; i < elev; ++i) {
    Kokkos::parallel_for("forward_sweep", Kokkos::RangePolicy<>(startidx, endidx), KOKKOS_LAMBDA (const int jc) {
      T m = 1.0 / (b_view(jc, i) - c_p(jc, i - 1) * a_view(jc, i));
      c_p(jc, i) = c_view(jc, i) * m;
      d_p(jc, i) = (d_view(jc, i) - d_p(jc, i - 1) * a_view(jc, i)) * m;
    });
    Kokkos::fence();
  }

  // Initialize the output array at the last level (elev-1)
  Kokkos::parallel_for("init_varout", Kokkos::RangePolicy<>(startidx, endidx), KOKKOS_LAMBDA (const int jc) {
    varout_view(jc, elev-1) = d_p(jc, elev-1);
  });
  Kokkos::fence();

  // Back substitution: update varout for columns from elev-2 down to slev.
  for (int i = elev - 2; i >= slev; --i) {
    Kokkos::parallel_for("back_substitution", Kokkos::RangePolicy<>(startidx, endidx), KOKKOS_LAMBDA (const int jc) {
      varout_view(jc, i) = d_p(jc, i) - c_p(jc, i) * varout_view(jc, i + 1);
    });
    Kokkos::fence();
  }

  c_p = Kokkos::View<T**>();
  d_p = Kokkos::View<T**>();
  // End timing and print the elapsed time
  auto end_time = std::chrono::high_resolution_clock::now();
  std::chrono::duration<double> elapsed_time = end_time - start_time;
  std::cout << "Elapsed time for tdma_solver_vec (Kokkos): " << elapsed_time.count() << " seconds" << std::endl;
}

template
void tdma_solver_vec<double>(const double* a, const double* b, const double* c, const double* d,
  int slev, int elev, int startidx, int endidx,
  int nrows, int ncols, double* varout);

template
void tdma_solver_vec<float>(const float* a, const float* b, const float* c, const float* d,
  int slev, int elev, int startidx, int endidx,
  int nrows, int ncols, float* varout);

